import pandas as pd
import numpy as np

from margot.data.features import BaseFeature


class SimpleReturns(BaseFeature):
    """Simple returns are the percent change from yesterdays close to today's close.

    Args:
        column (str): The name of a price time series.
    """

    label = 'simple_returns'

    def feature(self, series):
        return series.pct_change().fillna(0)


class LogReturns(BaseFeature):
    """Log returns can be summed over time.

    Args:
        column (str): The name of the price time series.
    """

    label = 'log_returns'

    def feature(self, series):
        return np.log(1 + series.pct_change().fillna(0))


class RealisedVolatility(BaseFeature):
    """Realised volatility measures the variability of returns over a lookback window.

    Args:
        column (str): The name of a returns time series.
        window (int): Lookback window in trading days.

    Raises:
        AttributeError: A lookback window is required.
    """

    label = 'realised_vol'
    window = None

    def feature(self, series):
        if not self.window:
            raise AttributeError(
                'you must supply a lookback window for RealisedVolatility')
        return series.rolling(
            self.window).std() * np.sqrt(252)


class SimpleMovingAverage(BaseFeature):
    """Simple moving average of lookback window.

    Args:
        column (str): The name of a returns time series.
        window (int): Lookback window in trading days.
    """

    window = None

    def get_label(self):
        return 'sma{}'.format(self.window)

    def feature(self, series):
        if not self.window:
            raise AttributeError(
                'you must supply a lookback window for SimpleMovingAverage')
        return series.rolling(self.window).mean()


class UpperBollingerBand(BaseFeature):
    """Upper bollinger band with window and standard deviation.

    Args:
        column (str): The name of a returns time series.
        window (int): lookback in trading days. Defaults to 20
        width (float): width in standard deviations. Defaults to 2.0
    """

    window = 20
    width = 2.0
    label = 'upper_boll_band'

    def feature(self, series):
        return series.rolling(self.window).mean(
        ) + series.rolling(self.window).mean().std() * self.width


class LowerBollingerBand(BaseFeature):
    """Lower bollinger band of window and standard deviation.

    Args:
        column (str): The name of a returns time series.
        window (int): lookback in trading days. Defaults to 20
        width (float): width in standard deviations. Defaults to 2.0
    """

    window = 20
    width = 2.0
    label = 'lower_boll_band'

    def feature(self, series):
        return series.rolling(self.window).mean(
        ) - series.rolling(self.window).mean().std() * self.width



from margot.signals.algos import BaseAlgo


class Position(object):
    """Represents a Position with a symbol and a weight.

    Arguments:
        symbol (str): The identifier of the symbol. e.g. 'SPY'.
        weight (float): A value between -1.0 and +1.0 representing
                        the weight of this symbol in the position list.
    """

    def __init__(self, symbol: str, weight: float):  # noqa: D107
        self.symbol = symbol
        self.weight = weight
        if self.weight > 1.0 or self.weight < -1.0:
            raise ValueError('weight must be a value between -1.0 and +1.0')

    def __repr__(self):
        """Represent."""
        return str((self.symbol, self.weight))



import pandas as pd


class Ratio(object):
    """Ratio of two series.

    Ratio = numerator / denomnator of two series.

    Args:
        numerator (str): the series to numerate the ratio
        denominator (str): the series to denominate the ratio
        label (str): give it a name for your dataframe. e.g. current_ratio
    """

    def __init__(self, numerator, denominator, label, **kwargs):    # noqa: D107
        self.numerator = numerator
        self.denominator = denominator
        self.series = None
        self.label = label
        self.__dict__.update(kwargs)

        if not self.numerator and self.denominator:
            raise AttributeError(
                "You must supply both a numerator and denominator")
        self.series = self.numerator.get_series().divide(self.denominator.get_series())

    def get_series(self):
        return self.series.rename(self.label)

    def to_pandas(self):
        return self.get_series().to_frame()

    @property
    def latest(self):
        """Return the latest value in this series"""
        return self.get_series().tail(1)[0]



import os


def test_symbol():
    from margot.data import Symbol
    from margot.data.column import alphavantage as av
    from margot.data.features import finance

    class Equity(Symbol):
        adjusted_close = av.Column(
            time_series='adjusted_close')
        volume = av.Column(
            time_series='volume')

        simple_returns = finance.SimpleReturns(column='adjusted_close')
        log_returns = finance.LogReturns(column='adjusted_close')
        realised_vol = finance.RealisedVolatility(
            column='log_returns', window=30)

        upper_band = finance.UpperBollingerBand(
            column='adjusted_close', window=20, width=2.0)
        sma20 = finance.SimpleMovingAverage(column='adjusted_close', window=20)
        lower_band = finance.LowerBollingerBand(
            column='adjusted_close', window=20, width=2.0)

    env = {'DATA_CACHE': os.path.join(os.getcwd(), 'data')}
    spy = Equity(symbol='SPY', trading_calendar='NYSE', env=env)


def test_frame():
    from margot.data import MargotDataFrame, Symbol, Ratio
    from margot.data.column import cboe
    from margot.data.features import finance

    class Index(Symbol):
        close = cboe.Column(time_series='close')

    class VXBasis(MargotDataFrame):
        vixm = Index(symbol='VIX3M', trading_calendar='NYSE')
        vix3m = Index(symbol='VIX', trading_calendar='NYSE')
        vx_basis = Ratio(
            numerator=vixm.close,
            denominator=vix3m.close,
            label='vx_basis_ratio')

    vxbasis = VXBasis()
    vxbasis.to_pandas()
    vxbasis.vx_basis.latest
    vxbasis.vx_basis.to_pandas()


def test_constructors():
    from margot.data import MargotDataFrame, Symbol, Ratio
    from margot.data.column import alphavantage as av

    class Index(Symbol):
        adjusted_close = av.Column(
            time_series='adjusted_close')

    spy = Index(symbol='SPY', trading_calendar='NYSE')
    vtwo = Index(symbol='VTWO', trading_calendar='NYSE')

    assert(
        spy.to_pandas().tail()[
            'SPY', 'adjusted_close'].sum() != vtwo.to_pandas().tail()[
            'VTWO', 'adjusted_close'].sum()
    )


def test_finance_features():
    from margot.data import MargotDataFrame, Symbol, Ratio
    from margot.data.column import alphavantage as av
    from margot.data.features import finance

    class Index(Symbol):
        adjusted_close = av.Column(
            time_series='adjusted_close')

        simple_returns = finance.SimpleReturns(column='adjusted_close')
        log_returns = finance.LogReturns(column='adjusted_close')
        real_vol = finance.RealisedVolatility(column='log_returns', window=20)
        sma = finance.SimpleMovingAverage(column='adjusted_close', window=10)
        upper = finance.UpperBollingerBand(
            column='adjusted_close', window=20, width=2.0)
        lower = finance.LowerBollingerBand(
            column='adjusted_close', window=20, width=2.0)

    spy = Index(symbol='SPY', trading_calendar='NYSE')

    spy.to_pandas()



from inspect import getmembers

import pandas as pd

from trading_calendars import get_calendar

from margot.data.column import BaseColumn
from margot.data.features import BaseFeature
from margot.data.ratio import Ratio


class Symbol(object):
    """A Symbol, represents a securitised tradable asset.

    A symbol can contain columns and features as members.

        Usage example:

        class Equity(Symbol):
            close = av.Column(time_series='adjusted_close')

        spy = Equity(symbol='SPY', trading_calendar='NYSE') 

    Args:
        symbol (str): the code for this symbol
        trading_calendar (str): ISO Code market identifier
            uses https://github.com/quantopian/trading_calendars/blob/master/README.md
        env (dict): Optional - pass in env values rather than use sysenv.

    """

    def __init__(self, symbol: str, trading_calendar: str, env: dict = {}):    # noqa: D107
        self.symbol = symbol
        self.env = env
        self.trading_calendar = get_calendar(trading_calendar)

        self.columns = [
            member for member,
            ref in getmembers(self, lambda m: isinstance(m, BaseColumn))]

        self.features = [
            member for member,
            ref in getmembers(self, lambda m: isinstance(m, BaseFeature))]

        self.ratios = [
            member for member,
            ref in getmembers(self, lambda m: isinstance(m, Ratio))]

        for col in self.columns:
            new_col = getattr(self, col).clone()
            setattr(self, col, new_col)
            getattr(self, col).setup(symbol=symbol, env=self.env)

        for feature in self.features:
            base_series_name = getattr(self, feature).get_column_name()
            new_feat = getattr(self, feature).clone()
            setattr(self, feature, new_feat)
            base_col = getattr(self, base_series_name)
            getattr(self, feature).set_column(base_col)

        # TODO ratios

        super().__init__()

    def to_dict(self):
        elements = self.columns + self.features + self.ratios
        return {(self.symbol, elt): getattr(self, elt).get_series()
                for elt in elements}

    def to_pandas(self):
        return pd.DataFrame(self.to_dict())

    def refresh(self):
        """Refresh all columns in this Symbol."""
        [getattr(self, member).refresh() for member in self.columns]

    def simulate(self, when):
        """Make the Symbol simulate a datetime in history.

        Used for backtesting to simplify the writing of trading
            algorithms.

        Args:
            when (tz_aware datetime or pd.Timestamp): when to go back to.
        """
        for col in self.columns:
            getattr(self, col).simulate(when)



import os

import pandas as pd

from margot.data.column import BaseColumn


class Column(BaseColumn):
    """A single OHLC timeiseries from CBOE.

    Currently supports the symbols, 'VIX' and 'VIX3M'.

    Example:
         = cboe.Column(time_series='close')

    Args:
        column (str): the name of the column that will be returned
    """

    INDEX = {
        'VIX': {
            'url': 'http://www.cboe.com/publish/scheduledtask/mktdata/datahouse/vixcurrent.csv',
            'skiprows': 1,
            'index_col': 0
        },

        'VIX3M': {
            'url': 'http://www.cboe.com/publish/scheduledtask/mktdata/datahouse/vix3mdailyprices.csv',
            'skiprows': 2,
            'index_col': 0
        }
    }

    def clean(self, df):
        """Additional cleaning for CBOE symbols."""
        df = df.rename(mapper={
            'VIX Open': 'open',
            'VIX High': 'high',
            'VIX Low': 'low',
            'VIX Close': 'close',
            'OPEN': 'open',
            'HIGH': 'high',
            'LOW': 'low',
            'CLOSE': 'close'
        }, axis='columns')

        return super().clean(df)

    def fetch(self, symbol: str):
        """Fetch from remote - this could be the only service specific thing."""
        print('fetching ({})'.format(symbol))

        try:
            df = pd.read_csv(self.INDEX[symbol].get('url'),
                             skiprows=self.INDEX[symbol].get('skiprows'),
                             index_col=self.INDEX[symbol].get('index_col'),
                             parse_dates=True)
        except KeyError:
            raise KeyError('The CBOE fetcher doesn\'t know about that symbol.')

        return self.clean(df)



import os

from alpha_vantage.timeseries import TimeSeries

from margot.data.column import BaseColumn


class Column(BaseColumn):
    """A single Symbol time series from AlphaVantage.

    Example:
        volume = av.Column(function='historical_daily_adjusted', time_series='volume')

    Args:
        time_series (str): the name of the time-series that will be returned
    """

    def clean(self, df):
        """Clean the df."""
        # Standardise the column names
        df = df.rename(mapper={
            '1. open': 'open',
            '2. high': 'high',
            '3. low': 'low',
            '4. close': 'close',
            '5. adjusted close': 'adjusted_close',
            '6. volume': 'volume',
            '7. dividend amount': 'divident_amount',
            '8. split coefficient': 'split_coefficient'
        }, axis='columns')
        return super().clean(df)

    def fetch(self, symbol: str):
        """Fetch from remote - this could be the only service specific thing."""
        print('fetching ({})'.format(symbol))
        ts = TimeSeries(
            key=self.env.get(
                'ALPHAVANTAGE_API_KEY',
                os.environ.get('ALPHAVANTAGE_API_KEY')),
            output_format='pandas')
        df, metadata = ts.get_daily_adjusted(symbol, outputsize='compact')
        return self.clean(df)



import pandas as pd

from margot.data.column import BaseColumn


class BaseFeature(object):

    INITED = False

    def get_label(self):
        try:
            return self.label
        except AttributeError:
            raise AttributeError(
                'Please declare a label attribute for this feature')

    def __init__(self, column: str, *args, **kwargs):  # noqa: D107
        self.column = column
        self.__dict__.update(kwargs)
        self.kwargs = kwargs
        self.args = args
        self.series = None
        self.base_column = None

    def clone(self):
        """Return a new instance of oneself."""
        return self.__class__(self.column, *self.args, **self.kwargs)

    def set_column(self, base_column: BaseColumn):
        self.base_column = base_column

    def get_column_name(self) -> str:
        return self.column

    def get_series(self):
        if self.series is None:
            # TODO - consider set_column - used after cloning.
            series = self.feature(self.base_column.get_series())
            self.series = series.rename(self.get_label())
        return self.series

    def get_label(self):
        return self.label

    def feature(self, series: pd.Series):
        raise NotImplementedError("please implement the feature")

    @property
    def latest(self):
        """Return the latest value in this series"""
        return self.get_series().tail(1)[0]




def test_algo():
    from margot.signals import algos
    from margot.data import MargotDataFrame, Symbol, Ratio
    from margot.data.column import cboe

    class Index(Symbol):
        close = cboe.Column(time_series='close')

    class VXBasis(MargotDataFrame):
        vixm = Index(symbol='VIX', trading_calendar='NYSE')

    class MyAlgo(algos.BaseAlgo):
        data = MargotDataFrame()

    myalgo = MyAlgo(env={})



from inspect import getmembers
from datetime import datetime

import pandas as pd

from margot.data.column import BaseColumn
from margot.data.features import BaseFeature
from margot.data.symbols import Symbol
from margot.data.ratio import Ratio


class MargotDataFrame(object):
    """An MargotDataFrame brings together symbols, columns, features and ratios.

    Args:
        env (dict): optional env dictionary as an alternative to sysenv variables.

    Raises:
        NotImplementedError: [description]

    Returns:
        [type]: [description]
    """

    def __init__(self, env: dict = {}):   # noqa: D107
        self.env = env

        self.symbols = [
            name for name,
            ref in getmembers(self, lambda m: isinstance(m, Symbol))]

        self.features = [
            name for name,
            ref in getmembers(self, lambda m: isinstance(m, BaseFeature))]

        self.ratios = [
            name for name,
            ref in getmembers(self, lambda m: isinstance(m, Ratio))]
        super().__init__()

    def to_pandas(self, when: datetime = None) -> pd.DataFrame:
        """Return a pandas Dataframe representing this MargotDataFrame.

        Args:
            when (datetime, optional): slice to only show data that was
                available at when.
                That is, the EOD from the previous day.
                Defaults to None.

        Returns:
            pd.DataFrame: a Pandas dataframe representing all data from
                the MargotDataFrame
        """
        # Get the elements one at a time, to pandas them and ensemble.
        if len(self.symbols) == 1:
            df1 = self.symbols[0].to_pandas()
        elif len(self.symbols) > 1:
            df1 = pd.concat([getattr(self, name).to_pandas()
                             for name in self.symbols], axis=1)
        else:
            df1 = pd.DataFrame()

        df2 = pd.DataFrame({('margot', name): getattr(self, name).get_series()
                            for name in self.ratios + self.features})

        df = pd.concat([df1, df2], axis=1)

        if when:
            df = df.to_pandas().shift()[:when]

        return df

    def refresh(self):
        """Refresh all Symbols in this DataFrame."""
        for member in self.symbols:
            getattr(self, member).refresh()

    @property
    def start_date(self):
        """First Timestamp of the time-series index.

        Returns:
            Timestamp: a pandas timestamp.
        """
        return self.to_pandas().index.min()

    @property
    def end_date(self):
        """Last Timestamp value of the time-series index.

        Returns:
            Timestamp: a pandas timestamp.
        """
        return self.to_pandas().index.max()

    @property
    def index(self):
        """Return the time-series index.

        Returns:
            pd.Index: a pandas timeseries index.
        """
        return self.to_pandas().index

    def simulate(self, when):
        """Create a dataframe simulating a datetime in history.

        Used for backtesting to simplify the writing of trading
            algorithms.

        After simulating a historical datetime, it is not possible to
            go back to the future.

        Args:
            when (tz_aware datetime or pd.Timestamp): when to go back to.
        """
        for symbol in self.symbols:
            getattr(self, symbol).simulate(when)

        for feature in self.features + self.ratios:
            getattr(self, symbol).recalc()



import os
import logging
import itertools
from datetime import date, datetime, timedelta
from copy import deepcopy

import pandas as pd
import numpy as np
from alpha_vantage.timeseries import TimeSeries
from trading_calendars import get_calendar

import pytz
from margot.signals.periods import DAILY
from margot.data import MargotDataFrame


class BaseAlgo(object):
    """A base class to inherit when implementing your trading algorithm.

    You should at least implement signal() which is the output
    of a trading algorithm.

    Args:
        env (dict): a dictionary of environment variables, e.g. API keys.
                    Overrides anything provided in sys env.

    Raises:
        ValueError: the attribute, 'data' must be a reference to a MargotDataFrame.
        NotImplementedError: If your subclass does not implement signal(), you will
            receive a NotImplementedError.

    """

    frequency = DAILY
    data = None

    def __init__(self, env: dict = {}, calendar='XNYS'):  # noqa: D107
        self.env = env
        self.calendar = calendar
        if not isinstance(self.data, MargotDataFrame):
            raise ValueError('Please set data to reference a MargotDataFrame')

    def signal(self) -> list:
        """Return a list of Position objects for a given datetime."""
        raise NotImplementedError("You must implement signal")

    def simulate(self, when: datetime):
        """Simulate a signal from a point in time.

        Args:
            when (datetime): when to go back to

        Returns:
            pd.DataFrame: time-series of Positions.
        """
        self.original_data = deepcopy(self.data)
        self.data.simulate(when)
        positions = self.signal()
        self.data = self.original_data
        return positions

    def run(self, when: datetime = None):
        """Call to run this algo at a point in time (when).

        We store the results of fetch_data() so that we can backtest
        without hitting the data provider more than once.

        :return: a list of Position objects for a given datetime
        """

        if not when:
            when = datetime.now(tz=pytz.UTC)

        return self.signal()



import os
import logging
from pathlib import Path
import pytz
import pandas as pd


class BaseColumn(object):
    """A Column represents a single time series of a symbol.

    This could be adjusted_close, open, volume - etc.

    Example:
        volume = column.AlphaVantage(function='historical_daily_adjusted', field='adjusted_close')

    Args:
        function (str): the name of the function passed to the Alphavantage API
        time_series (str): the name of the time_series that will be returned
    """

    INITED = False

    def __init__(self, time_series: str, *args, **kwargs):  # noqa: D107
        self.time_series = time_series
        self.series = None

    def get_label(self):
        """Return the label for this column."""
        return self.series.name

    def clone(self):
        """Return a new instance of oneself."""
        return self.__class__(self.time_series)

    def setup(self, symbol: str, env: dict = {}):
        """Setup the column.

        Called by the Symbol so that the symbol name can be passed.
        """
        self.symbol = symbol
        self.env = env

        # TODO File names should be managed in a central configuration
        data_cache = env.get('DATA_CACHE', os.environ.get('DATA_CACHE'))
        Path(data_cache).mkdir(parents=True, exist_ok=True)

        self.hdf5_file = os.path.join(
            data_cache, '{}.hdf5'.format(self.symbol))

    def clean(self, df):
        """Clean the data."""
        df = df.sort_index()
        # make tz aware if not already
        if not isinstance(df.index.dtype, pd.DatetimeTZDtype):
            df = df.tz_localize(pytz.UTC)

        return df

    def load_or_fetch_series(self, symbol: str):
        """Load of fetch the Dataframe, return the series.

        In order to return the time-series, first determine if we
        have it and can return it, or if we need to fetch it.

        TODO: Test for up-to-dateness (or maybe that happens in Symbol)?

        Args:
            symbol (str): the name of the symbol to fetch.

        Returns:
            pd.Series: time-series of the column
        """
        try:
            df = self.load(symbol)
        except (KeyError, FileNotFoundError):
            df = self.refresh()
        return df[self.time_series]

    def refresh(self):
        """Refresh the data from the source.

        Returns:
            pd.DataFrame: the whole dataframe (cleaned)
        """
        df = self.fetch(self.symbol)
        df = self.clean(df)
        self.save(df, self.symbol)
        return df

    def fetch(self, symbol: str):
        raise NotImplementedError(
            'This is implementation specific to the data provider.')

    def load(self, symbol: str):
        """Load it."""
        return pd.read_hdf(
            self.hdf5_file,
            key=symbol)

    def save(self, df, symbol):
        """Save it."""
        df.to_hdf(self.hdf5_file, key=symbol)

    def get_series(self, when = None):
        """Get the data series as a pandas series.

        Args:
            when (datetime): 

        Returns:
            pd.Series: time series of the field
        """
        if self.series is None:
            self.series = self.load_or_fetch_series(self.symbol)

        self.series = self.series[:when]

        self.INITED = True
        return self.series

    @property
    def latest(self):
        """Return the latest value in this series."""
        return self.get_series().tail(1)[0]



import os
import logging
from datetime import date, datetime, timedelta

import pandas as pd
import numpy as np

import pytz


class BackTest(object):
    """Backtest an trading algo that's a descendent of BaseAlgo.

    Longer class information....
    Longer class information....

    Attributes:
        algo: A boolean indicating if we like SPAM or not.
        starting_balance: An integer count of the eggs we have laid.
    """

    positions = None
    algo = None
    symbols = set()

    def __init__(self, algo_class, start_balance=100000):   # noqa: D107
        self.start_balance = start_balance
        self.algo_class = algo_class
        self.algo = algo_class()
        self.positions = None
        self.trades = None
        self.returns = pd.DataFrame(
            columns=[
                'returns',
                'log_returns'],
            dtype='float64')

    def calc_daily_returns(self, day):
        """Calculate the returns based on yesterdays positions, MoC to MoC.

        Args:
            day (date): the date for which we calculate returns.

        Daily returns are the difference between yesterdays adjusted_close
        and todays adjusted_close.
        """
        simple_returns = 0.0

        for symbol in self.symbols:
            # simple_returns = yesterdays_positions * todays_pct_change
            try:
                todays_pct_change = self.algo.symbols[symbol].data.pct_change(
                ).loc[day].adjusted_close
                yesterdays_position = self.positions.shift().loc[day, symbol]
                simple_returns = simple_returns + \
                    (todays_pct_change * yesterdays_position)
            except KeyError:
                pass
        self.positions.at[day, 'returns'] = simple_returns
        self.positions.at[day, 'log_returns'] = np.log(1 + simple_returns)

        return self.positions.dropna()

    def walk_forward(self, start: date, end: date):
        """Backtest the algo, walk forward for every trading day in the date range.

        Calculates the returns from taking the previous days positions.

        Args:
            start (date): The first day of the backtest.
            end (date): The last day of the backtest.

        Return:
            a DataFrame of daily simple returns, and log returns.

        Note: A walk forward backtest is much slower than backtesting a dataframe with shift(),
        but we can use it on the same algo that is deployed live. It's also a good way to standardise
        backtesting so that we can keep margot simple, and reuse performance calcs etc.
        """
        trading_days = get_calendar(
            self.algo.calendar).schedule[start:end].index

        for day in trading_days:
            # TODO simplify this.
            row = [p.as_dict() for p in self.algo.execute(day)]
            mapper = {k: v for d in row for k, v in d.items()}
            self.symbols = self.symbols.union(mapper.keys())
            self.positions = self.positions.append(
                pd.DataFrame(mapper, index=[day]))
            self.calc_daily_returns(day)
        return self.positions

    def create_position_timeseries(self):
        """Create Position time-series from signals.

        Returns:
            [type]: [description]
        """
        for day in self.algo.data.index:
            pass
        return None

    def run(self):
        """Run the backtest.

        Returns:
            [type]: [description]
        """
        # first calculate the positions time series
        self.positions = self.create_position_timeseries()
        # deduce the trades; simulate entry and exit prices.
        # calculate the returns.
        # caclulate lookback rolling volatility
        # simulate resizing at trading time, according to a target volatility
        # simulate volatility sized returns.
        return ('returns', 'trades')
